from typing import Any,Dict
from collections import namedtuple
import math
import numpy as np
import sasktran as sk
from skimpy import Instrument, ForwardModel

#------------------------------------------------------------------------------
#               encode_aim_north_level1_data(
#------------------------------------------------------------------------------

def encode_aim_north_level1_data(   instrument  : Instrument,                       # Radiance information for each observation set sample is stored in the indiviual pixels
                                    forwardmodel: ForwardModel,                     # Forward model used to calculate the radiance,
                                    center_pos  : np.ndarray,                       # GEO Position of platform for each sample Array (3, Nt)
                                    center_los  : np.ndarray,                       # GEO central look vector of each pixel in each sample, Array(3, Numpix, Nt)
                                    utc         : np.ndarray,                       # UTC time of each observation, Array(Nt) of np.datetime64 or possibly float if MJD
                                    center_wave : np.ndarray )->Dict[str,Any]:    # Central wavelength of each pixel in each sample, Array(Numpix,Nt)
    """
    Function encodes the radiance signals into Aim North Level 1 data. This is typically invoked as a callback
    function near the end of instrument.calculate_radiance(). Its purpose is to return a single python object that
    has the Level 1 data for further simulation/retrieval purposes.
    """

    skengine   = forwardmodel.radiance_simulator().engine()
    skgeodetic = sk.Geodetic()
    ok,sun     =  skengine._iskengine.GetProperty('sun')
    refpoint   =  skengine._iskengine.GetProperty('referencepoint')

    fer      = instrument.front_end_radiance()
    numwave  = instrument.front_end_radiance().num_pixels()                          # get the number of wavelengths/pixels
    nt       = center_los.shape[2]                                                   # get the number of time samples
    level1_radiance   = np.zeros( (numwave, nt) )
    level1_irradiance = np.zeros( (numwave, nt) )
    level1_toaalbedo  = np.zeros( (numwave, nt) )
    level1_photonflux = np.zeros( (numwave, nt) )
    for iw in range( fer.num_pixels()):
        level1_radiance  [iw, :] = fer.pixels[iw].radiance
        level1_irradiance[iw, :] = fer.pixels[iw].irradiance
        level1_toaalbedo [iw, :] = fer.pixels[iw].toa_albedo
        level1_photonflux[iw, :] = fer.pixels[iw].photon_flux

    satellite_geocentric_pos = np.reshape( center_pos[:,:], (3, nt) )
    satellite_geocentric_los = np.reshape( center_los[:,0,:], (3, nt) )
    satellite_geodetic_pos   = np.zeros( (3,  nt) )
    ground_geodetic_pos      = np.zeros( (3,  nt) )
    ground_sza               = np.zeros( (nt,) )
    ground_ssa               = np.zeros( (nt,) )
    ground_saa               = np.zeros( (nt,) )
    look_zenith              = np.zeros( (nt,) )
    wavelens                 = np.reshape   ( center_wave[:,0],  (numwave,) )

    for it in range(nt):
        skgeodetic.from_xyz( satellite_geocentric_pos[:,it] )             # Set the satellite position for this sample
        satellite_geodetic_pos[0,it] = skgeodetic.latitude                #
        satellite_geodetic_pos[1,it] = skgeodetic.longitude
        satellite_geodetic_pos[2,it] = skgeodetic.altitude

        intercepts = skgeodetic.altitude_intercepts( 0.0,  satellite_geocentric_pos[:,it], satellite_geocentric_los[:,it])
        ground_point = intercepts[0]
        skgeodetic.from_xyz( ground_point )

        ground_geodetic_pos[0, it] = skgeodetic.latitude
        ground_geodetic_pos[1, it] = skgeodetic.longitude
        ground_geodetic_pos[2, it] = skgeodetic.altitude
        up     = skgeodetic.local_up
        south  = skgeodetic.local_south
        west   = skgeodetic.local_west
        z      = np.dot( up, sun)
        x      = np.dot( sun, -south)
        y      = np.dot( sun, -west )
        coslookzenith = np.dot( -up, satellite_geocentric_los[:,it] )
        cosssa = np.dot( satellite_geocentric_los[:,it], sun )
        ground_saa[it] = math.degrees( math.atan2(y,x) )
        ground_sza[it] = math.degrees( math.acos(z) )
        ground_ssa[it] = math.degrees( math.acos(cosssa) )
        look_zenith[it]= math.degrees( math.acos(coslookzenith) )

    level1 ={ 'version'                    : '0.1',
              'radiance'                   : level1_radiance,
              'irradiance'                 : level1_irradiance,
              'toa_albedo'                 : level1_toaalbedo,
              'photon_flux'                : level1_photonflux,
              'wavelengths'                : wavelens,
              'utc'                        : utc[0],
              'satellite_geocentric_pos'   : satellite_geocentric_pos,
              'satellite_geocentric_los'   : satellite_geocentric_los,
              'satellite_geodetic_pos'     : satellite_geodetic_pos,
              'ground_points_geodetic_pos' : ground_geodetic_pos,
              'sun_vector'                 : sun,
              'ground_sza'                 : ground_sza,
              'ground_ssa'                 : ground_ssa,
              'ground_saa'                 : ground_saa,
              'look_zenith'                : look_zenith,
              'no2_profile'                : 'THIS WILL BE THE NO2 PROFILE'
              }

    return level1


